--------------------------------------------------------------------------
-- - - - - - - - - - - - - - - Dungeon object - - - - - - - - - - - - - - - 
---------------------------------------------------------------------------

-- Dungeon objects have several levels (consisting of Level objects) which
-- together represent a whole dungeon.

rawset(_G, "Dungeon", {nrOfLevels, height, width, levels})
Dungeon.__index = Dungeon

function Dungeon:new(nrOfLevels, height, width)
	if (nrOfLevels < 1) then error("Cannot create dungeon! Number of levels value cannot be less than 1!") end
	local dungeon = {}
	dungeon.nrOfLevels = nrOfLevels
	dungeon.height = height
	dungeon.width = width
	dungeon.levels = {}

	setmetatable(dungeon, Dungeon)
	return dungeon
end

function Dungeon:generateDungeon()
	local ldinc = (self.nrOfLevels/5) or 1 -- Dungeon Difficulty Incrementor
	-- This tells us how many floors it takes to increment the dungeon difficulty by 1
	-- Ex. a 15 floor dungeon would increment every 3 levels since there are 5 difficulty levels.
	local ld = 1 -- Dungeon Difficulty
	for i=1,self.nrOfLevels do
		local newLevel = Level:new(self.height, self.width)
		newLevel:generateLevel()
		self.levels[i] = newLevel
		
		newLevel.difficulty = ld
		--print("Level "..i.." assigned difficulty of: "..ld)
		if not (i%ldinc) and (ld < 5)
			ld = $ + 1
		end
	end
end

function Dungeon:initPlayerSpawn(level)
	local c=level:getRoot().center
	local adj=getAdjacentPos(c[1], c[2])
	local i=1
	local endr, endc
	repeat
		endr, endc = adj[i][1], adj[i][2]
		i = $ + 1
	until (level:getTile(endr,endc).class == Tile.WALK)

	level:getTile(endr,endc).class = Tile.SPAWN
	
	-- Let's Be generous and surround this tile by a bunch of ring tiles.
	local adj = getAdjacentPos(endr,endc) -- 3x3 square
	-- Build around this floor.
	for i=1,#adj do
		endr, endc = adj[i][1], adj[i][2]
		if level:getTile(endr,endc)
		and not (level:getTile(endr,endc).class == Tile.SPAWN) -- Do not touch the player spawn!
			if (i&1)
				level:getTile(endr, endc).class = Tile.RING -- Ring tile!
			else
				level:getTile(endr, endc).class = Tile.WALK -- Every other tile is a walkable Tile.
			end
		end
	end
end

function Dungeon:initEnd(level)
	local c=level:getEnd().center
	local adj=getAdjacentPos(c[1], c[2])
	local i=1
	local endr, endc
	repeat
		endr, endc = adj[i][1], adj[i][2]
		i = $ + 1
	until (level:getTile(endr,endc).class == Tile.WALK)

	level:getTile(endr,endc).class = Tile.BOSS
	level:getEnd().hasBoss = true
end

function Dungeon:initSprings(level)
	level:addSprings(1)
end

function Dungeon:initKeys(level)
	if P_RandomChance((FRACUNIT*3)/20) -- 15% chance of a key floor
		local room 
		if (#level.rooms > 1)
			repeat
				room = level:getRandRoom(2, #level.rooms) -- Do NOT place key in root room.
			until (room.hasSpring ~= true) -- Do NOT place key in the same room as spring.
		else -- Failsafe
			room = level:getRoot()
		end
		
		local corners = room:getCorners()
		local start, endc, endr = corners[1], corners[2], corners[3]
		local row, col
		local i = 5 -- Hard breakout
		
		repeat
			-- Choose tile starting from top-left
			row = P_RandomRange(start[1], endr[1])
			col = P_RandomRange(start[2], endc[2])
			i = $ - 1
		until (level:getTile(row,col).class == Tile.WALK) or (i <= 0)
		
		level:getTile(row,col).class = Tile.KEY
		level.hasKey = 1
		--print("Level was assigned a Key!")
	--else
		--print("-")
	end
end

/*function Dungeon:printDungeon(level)
	if not level -- No level?
		-- Print every level``
		for i=1,#self.levels do
			local s = "L E V E L  "..i
			print(s)
			self.levels[i]:printLevel()
			print()
		end
	else
		if not tonumber(level) then error("Desired input is invalid") end
		if (level > self.nrOfLevels) then error("Desired level cannot be printed: Doesn't exist.") end

		local s = "L E V E L  "..level
		print(s)
		self.levels[level]:printLevel()
		print()
	end
end*/