local insert = table.insert

---------------------------------------------------------------------------------------
-- - - - - - - - - - - - - - - - - - - Level object - - - - - - - - - - - - - - -- - --
---------------------------------------------------------------------------------------

-- A Level object consist of several Tile objects which together make up 
-- one dungeon level.

rawset(_G, "Level", { height, width, matrix, rooms, springs})
Level.__index = Level

function Level:new(height, width)
	if height > 47 or width > 47 then error("Level must NOT have a height greater than 47") end

	local level = {
		height=height,
		width=width,
		matrix={},
		rooms = {},
		springs = {},
		rootRoom=nil,
		endRoom=nil,
		hasKey=0, -- 0 for false, 1 for true.
		difficulty = 1,
		minRoomSize = 3,
		maxRoomSize = 8,
		maxRooms = 6
	}

	setmetatable(level, Level)
	return level
end

-- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- 

function Level:generateLevel()
	-- A default generation of a level including rooms, corridors to each room forming and Pits.
	-- Just create our floors for now. Objects will be spawned via Dungeon level. Not by Level level.
	self:initMap()
	self:generateRooms()
	local root=self:getRoomTree()
	self:buildCorridors(root)
	
	if (#self.rooms > 1)
		self:addCycles(P_RandomRange(1,#self.rooms)) -- Add extra cooridors between rooms.
	end

	self:addItems(P_RandomRange(0,#self.rooms/2))
	self:addEtc()
end

-- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- 

function Level:initMap()
	-- Create void
	for i=0,self.height do
		self.matrix[i] = {}
		for j=0,self.width do
			self.matrix[i][j] = Tile:new(Tile.WALL) -- Everything is wall.
		end
	end

	-- Everything else is already a wall. Not needed. Leftover code.
	--self:addWalls(0, 0, self.height, self.width) -- Add walls to the sides of the map.
end 

-- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- 

function Level:printLevel()
	for i=0,self.height do
		local row=""
		for j=0,self.width do
			row=row..self.matrix[i][j].class..","
			--row=row.."("..i..","..j..")"..","
			--row=row..self.matrix[i][j].roomId..","  -- for exposing room-ids
			--row=row..self.matrix[i][j].enemyId..","  -- for exposing enemytype-ids
		end
		print(row)
	end
end

-- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- 

function Level:getRandRoom(a, b)
	-- return: Random room in level
	local i

	if a and b
		i = P_RandomRange(a,b)
	else
		i = P_RandomRange(1,#self.rooms)
	end
	return self.rooms[i]
end

-- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- 

function Level:getRoot()
  -- return: Room that is root of room tree if such has been generated.
  return self.rootRoom
end

function Level:getEnd()
  -- return: Leaf room added last to tree if such has been generated.
  return self.endRoom
end

-- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- 

function Level:getTile(r, c)
	return self.matrix[r][c]
end

function Level:getAdjacentTiles(row, col)
	-- returns table containing all adjacent tiles to given position.
	-- Including self!

	local result={}
	local adj=getAdjacentPos(row,col)
	
	for i=1,#adj do
		local row, col = adj[i][1], adj[i][2]
		insert(result, self:getTile(row, col))
	end
	return result
end

function Level:buildTile(r, c)
	-- Builds floor tile surrounded by other tiles
	local adj = getAdjacentPos(r,c) -- 3x3 square
	
	self:getTile(r, c).class = Tile.WALK
	-- Build around this floor.
	for i=1,#adj do
		r, c = adj[i][1], adj[i][2]
		if self:getTile(r,c)
		and not (self:getTile(r,c).class == Tile.WALK) -- Do not touch already made floors
		--and not (self:getTile(r,c).class == Tile.CORR)  -- Do not touch already made Cooridors
		and not self:getTile(r,c).roomId -- Do not intrude on rooms!!
			self:getTile(r, c).class = Tile.WALK
		end
	end
end

-- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- 

function Level:getSprings()
	-- Index is [1] for row, [2] for col on individual spring.
	return self.springs
end

function Level:addSprings(maxSprings)
	-- Adds ascending springs to random rooms.
	-- Number of springs depend on number of rooms.
	local room = self:getRandRoom(2,#self.rooms)
	
	if (#self.rooms > 1)
		local springs = P_RandomRange(1,maxSprings)
		repeat
			if not room.hasSpring then
				self:placeSpring(room, springs)
				springs = springs - 1
			end
		until springs==0
	else
		-- One room? One spring.
		self:placeSpring(room, 1)
	end
end

function Level:placeSpring(room, springs)
	-- Places staircase in given room. 
	-- Position is random number of steps away from center.

	local steps = P_RandomRange(0, FixedFloor((self.maxRoomSize*FRACUNIT)/2)/FRACUNIT)
	local nrow, ncol = room.center[1], room.center[2]
	local row, col
	
	repeat
		row, col = nrow, ncol
		repeat
			nrow, ncol = getRandNeighbour(row, col)
		until self:getTile(nrow, ncol).class == Tile.WALK
		steps=steps-1
	until (self:getTile(nrow, ncol).roomId ~= room.id or steps <= 0)

	-- Since we're passing `springs` into our function,
	-- I could make acending and descending Springs in the future?
	-- if springs%2 == 0 -- Would indicate a descending spring.
	-- Otherwise, we're ascending
	self:getTile(row, col).class = Tile.SPRING
	room.hasSpring = true
	insert( self.springs, { row, col } )
end

-- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- 

function Level:isRoom(row,col)
	return (self:getTile(row,col).roomId ~= 0)
end

function Level:buildRoom(x, y, w, h)
	-- init room object and paint room onto tiles.

	local id = #self.rooms+1
	local room = Room:new(id, x, y, w, h)
	local enemyId = P_RandomChance(FRACUNIT/2) and P_RandomRange(0,3) or 0 -- Give rooms enemy types!
	-- Convert to fixed point for a moment...
	--local fx, fy, fw, fh = x<<FRACBITS, y<<FRACBITS, w<<FRACBITS, h<<FRACBITS
	--local r,c = fw-(FixedFloor(fw - fx)/2)>>FRACBITS, 
	local r,c = w-(w-x)/2, h-(h-y)/2 -- Calculate the center of the room
	--print("Room "..id.." center is at ("..r..","..c..")")
	room:setCenter(r,c)
	insert(self.rooms, room)

	for i=x, w do
		for j=y, h do
			local tile = self:getTile(i,j)
			tile.roomId, tile.class, tile.enemyId = id, Tile.WALK, enemyId
		end
	end
	-- Everything else is a wall here. Not needed. Leftover code.
	--self:addWalls(x, y, w+1, h+1)
end
  
function Level:generateRooms() -- ROOMS
	for i = 1,self.maxRooms do
		self:generateRoom()
	end
end

function Level:generateRoom() -- ROOM
	-- Will randomly place rooms across tiles (no overlapping).

	local startRow = P_RandomRange(1, self.height-self.maxRoomSize)
	local startCol = P_RandomRange(1, self.width-self.maxRoomSize)

	local height = P_RandomRange(4, self.maxRoomSize)
	local width = P_RandomRange(4, self.maxRoomSize)

	for i=startRow, startRow+height do
		for j=startCol, startCol+width do

			if (self:isRoom(i,j)) then
				return -- Room is overlapping other room? Room is discarded
			end
		end
	end
	self:buildRoom(startRow, startCol, startRow+height, startCol+width)
end

-- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- 

function Level:getRoomTree()
	if #self.rooms < 1 then error("Can't generate room tree, no rooms exists") end

	-- Follow my lead here...
	-- This function calls to what's in LUA_MISC.
	-- The function in LUA_MISC makes a call to LUA_ROOM to ensure that the Tree Spanning can be set.
	--
	-- The cost of the spanning tree is the sum of the weights of all the edges in the tree.
	-- Prim’s Algorithm uses the Greedy approach to find the minimum spanning tree.
	-- For this algorithm, we grow the spanning tree from a starting position.
	-- 
	local root, lastLeaf = prims(clone(self.rooms))
	self.rootRoom = root
	self.endRoom = lastLeaf

	return root
end

-- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- 

function Level:buildCorridors(root)
	-- https://courses.cs.washington.edu/courses/cse326/03su/homework/hw3/dfs.html
	-- Recursive DFS function for building corridors to every neighbour of a room (root)

	for i=1,#root.neighbours do
		local neigh = root.neighbours[i]
		self:buildCorridor(root, neigh)
		self:buildCorridors(neigh) -- It may seem like we're calling an infinite loop, but we're not.
	end
end

function Level:buildCorridor(from, to)
	-- Parameters from and to are both Room-objects.

	local start, goal = from.center, to.center
	local srow, scol, grow, gcol = start[1], start[2], goal[1], goal[2]
	local diffr, diffc = (goal[1] - start[1]) >= 0 and 1 or -1, (goal[2] - start[2]) >= 0 and 1 or -1
	local random = P_RandomRange(0,1) -- false or true
	--print("Room "..from.id.." is attempting to reach room "..to.id)
	--print(diffr, diffc)
	
	if random > 0-- Random chance!
		--print("Random build 1!")
		-- Start with the col's (Left/right)
		for j = scol, gcol, diffc
			if not self:isRoom(start[1], j) -- Optimize rooms. :)
				self:getTile(start[1], j).class = Tile.CORR
			end
			if (j == gcol)
				 scol = j
			end
		end

		-- Then continue with the rows's (up/down)
		for k = srow, grow, diffr
			if not self:isRoom(k, scol)  -- Optimize rooms. :)
				self:getTile(k, scol).class = Tile.CORR
			end
		end
	else
		--print("Random build 2!")
		-- Start with the rows's (up/down)
		for k = srow, grow, diffr
			if not self:isRoom(k, start[2])  -- Optimize rooms. :)
				self:getTile(k, start[2]).class = Tile.CORR
			end
			if (k == grow)
				 srow = k
			end
		end
		
		-- Then continue with the col's (Left/right)
		for j = scol, gcol, diffc
			if not self:isRoom(srow, j)  -- Optimize rooms. :)
				self:getTile(srow, j).class = Tile.CORR
			end
		end
	end
	
	-- FUCK THIS ALGORITHM, I DON'T NEED nextTile.
	-- Besides, this errors out WAY too often to be reliable.
	/*local nextTile = findNext(start, goal) -- Go to LUA_MISC for this...
	repeat
		row, col = nextTile[1], nextTile[2]
		--print("(r,c): ("..nextTile[1]..", "..nextTile[2]..")")
		self:buildTile(row, col)

		if findNext(nextTile, goal)
			nextTile = findNext(nextTile, goal)
		else
			print("Breakout reached. (findNext = nil)")
			print("Room "..from.id.." attempted to reach room "..to.id)
			--print("Current (r,c): ("..row..", "..col..")")
			print("-----------------------------\n\n") -- Separator line
			break
		end
	until (self:getTile(nextTile[1], nextTile[2]).roomId == to.id)*/
end

function Level:addCycles(maxCycles)
  -- Adds corridors between random rooms.
	local from, to
	for i=1,maxCycles do
		--print("Adding cycle "..i.." between room(s)")
		from = self:getRandRoom()
		to = self:getRandRoom()
		if (from ~= to) -- The same room was picked! Don't build a cooridor, silly!
			self:buildCorridor(from, to)
		/*else
			print("Same rooms detected! Disregarding cycle #"..i)*/
		end
	end
end

-- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### --

function Level:addEtc()
	-- Iterates through the current level matrix to add pits/enemies/etc to cooridors and rooms.
	for i=0,self.height do
		for j=0,self.width do
			if (self:getTile(i,j).class == Tile.WALK)
			or (self:getTile(i,j).class == Tile.CORR)
				if P_RandomChance(FRACUNIT/15) -- 0.06 chance
					--self:buildPit(i,j)
					self:getTile(i, j).class = Tile.PIT
				elseif P_RandomChance(FRACUNIT/12) -- 0.8333 chance
					self:getTile(i, j).class = Tile.ENEMY
				elseif P_RandomChance(FRACUNIT/25) -- 0.05 chance
				and (self:getTile(i,j).class == Tile.WALK) -- Only apply to rooms
					self:getTile(i, j).class = Tile.RING
				end
			end
		end
	end
end

-- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### --

function Level:addItems(maxItems)
	if (maxItems > 0)
		for i=0, maxItems
			self:placeItems()
		end
	end
end

function Level:placeItems()
	-- Adds items to random room corners.
	local room, corners, rand
	local steps = 4 -- Number check
	local nrow, ncol
	
	repeat -- Pick a random corner that doesn't have an adjacent PIT
		room = self:getRandRoom(1,#self.rooms)
		corners = room:getCorners() -- Table
		rand = P_RandomRange(1,4)
		
		nrow, ncol = corners[rand][1], corners[rand][2]
		steps = steps - 1
	until (self:getTile(nrow, ncol).class ~= Tile.ITEM) -- Not already an item space
	/*and self:getTile(nrow+1, ncol+1) and (self:getTile(nrow+1, ncol+1).class ~= Tile.PIT) -- Check the 4 corners
	and self:getTile(nrow+1, ncol-1) and (self:getTile(nrow+1, ncol-1).class ~= Tile.PIT) 
	and self:getTile(nrow-1, ncol-1) and (self:getTile(nrow-1, ncol-1).class ~= Tile.PIT)
	and self:getTile(nrow-1, ncol+1) and (self:getTile(nrow-1, ncol+1).class ~= Tile.PIT)*/
	or (steps <= 0) -- Hard breakout check
	
	self:getTile(nrow, ncol).class = Tile.ITEM
end

function Level:buildPit(r, c)
	-- Builds pit tile surrounded by other tiles
	-- Similar to buildTile, but we're building a pit instead.
	local adj = getAdjacentPos(r,c) -- 3x3 square
	self:getTile(r, c).class = Tile.PIT
	-- Build around this floor and make pits bigger
	for i=1,#adj do
		r, c = adj[i][1], adj[i][2]
		if self:getTile(r,c)
		and (self:getTile(r,c):isWalkable()) -- Do not touch already made floors
		and not self:getTile(r,c).roomId -- Do not intrude on rooms!!
			self:getTile(r, c).class = Tile.PIT
		end
	end
end